import numpy

from judo.judo_backend import Backend, torch
from judo.judo_tensor import to_backend
from judo.typing import Scalar


class MetaTorchRandomState(type):
    @property
    def seed(cls):
        return torch.random.seed()


class TorchRandomState(metaclass=MetaTorchRandomState):
    def __init__(self, seed: Scalar):
        numpy.random.seed(seed)
        torch.random.manual_seed(int(seed))

    def __getattr__(self, item):
        return getattr(torch, item)

    @staticmethod
    def permutation(x):
        idx = torch.randperm(x.shape[0])
        sample = x[idx].to(Backend.get_device()).detach()
        return to_backend(sample)

    @staticmethod
    def random_sample(*args, **kwargs):
        sample = torch.rand(*args, **kwargs)
        return to_backend(sample)

    @staticmethod
    def choice(a, size=None, replace=True, p=None):
        a = to_backend(a)
        size = size if size is not None else 1
        if replace:
            size = size if isinstance(size, tuple) else (size,)
            indices = to_backend(torch.randint(len(a), size))
            samples = a[indices]
        else:
            indices = to_backend(torch.randperm(len(a)))[:size]
            samples = a[indices]
        return to_backend(samples)

    @staticmethod
    def uniform(
        low=0.0,
        high=1.0,
        size=None,
        dtype=None,
    ):
        uniform = torch.distributions.uniform.Uniform(low, high)
        if size is not None:
            size = size if isinstance(size, tuple) else (size,)
            sample = uniform.sample(size)
        else:
            sample = uniform.sample()
        if dtype is not None:
            sample = sample.to(dtype)
        return to_backend(sample)

    @staticmethod
    def randint(low, high, size=None, dtype=None):
        size = size if size is not None else (1,)
        size = size if isinstance(size, tuple) else (size,)
        data = torch.randint(low, high, size)
        if dtype is not None:
            data = data.to(dtype)
        return to_backend(data)

    @staticmethod
    def normal(loc=0, scale=1.0, size=None):
        size = size if size is not None else (1,)
        size = size if isinstance(size, tuple) else (size,)
        return to_backend(torch.normal(mean=loc, std=scale, size=size))

    @classmethod
    def random(cls, size=None):
        return cls.uniform(size=size)


class MetaRandomState(type):
    def __getattr__(cls, item):
        funcs = {
            "numpy": lambda x: getattr(cls._numpy_random_state, x),
            "torch": lambda x: getattr(cls._torch_random_state, x),
        }
        return Backend.execute(item, funcs)


class RandomState(metaclass=MetaRandomState):
    _random_seed = 160290
    _numpy_random_state = numpy.random.RandomState(seed=_random_seed)
    _torch_random_state = TorchRandomState(seed=_random_seed)

    @classmethod
    def seed(cls, seed: Scalar = _random_seed):
        cls._numpy_random_state = numpy.random.RandomState(seed=seed)
        cls._torch_random_state = TorchRandomState(seed=seed)


random_state = RandomState



import pytest

from judo import API, dtype, random_state


try:
    from PIL import Image

    PIL_MISSING = False
except ImportError:
    PIL_MISSING = True


class TestImages:
    @pytest.mark.skipif(PIL_MISSING, reason="PIL not installed")
    @pytest.mark.parametrize("size", [(64, 64, 3), (21, 73, 3), (70, 20, 3)])
    def test_resize_frame_rgb(self, size):
        random_state.seed(160290)
        frame = random_state.randint(0, 255, size=size, dtype=dtype.uint8)
        resized = API.resize_image(frame, width=size[1], height=size[0], mode="RGB")
        assert size == resized.shape

    @pytest.mark.skipif(PIL_MISSING, reason="PIL not installed")
    @pytest.mark.parametrize("size", [(64, 64), (21, 73), (70, 2)])
    def test_resize_frame_grayscale(self, size):
        random_state.seed(160290)
        frame = random_state.randint(0, 255, size=size + (3,), dtype=dtype.uint8)
        resized = API.resize_image(frame, size[1], size[0], "L")
        assert size == resized.shape



import uuid

import numpy
import xxhash

import judo
from judo.judo_backend import Backend


class Hasher:
    def __init__(self, seed: int = 0):
        self._seed = seed

    @property
    def uses_true_hash(self) -> bool:
        return Backend.use_true_hash()

    @staticmethod
    def hash_numpy(x: numpy.ndarray) -> int:
        """Return a value that uniquely identifies a numpy array."""
        # x = x.astype("|S576") if x.dtype == "O" else x
        return int(xxhash.xxh64_intdigest(x.tobytes()))

    @staticmethod
    def hash_torch(x):
        bytes = judo.to_numpy(x).tobytes()
        return int(hex(xxhash.xxh32_intdigest(bytes)))

    @staticmethod
    def get_one_id():
        return uuid.uuid1().int >> 64

    @classmethod
    def true_hash_tensor(cls, x):
        funcs = {
            "numpy": cls.hash_numpy,
            "torch": cls.hash_torch,
        }
        return Backend.execute(x, funcs)

    def hash_tensor(self, x):
        if self.uses_true_hash:
            x = x if judo.is_tensor(x) else judo.tensor([x])
            return self.true_hash_tensor(x)
        return self.get_one_id()

    def hash_iterable(self, x):
        hashes = [self.hash_tensor(xi) for xi in x]
        return judo.as_tensor(hashes, dtype=judo.dtype.hash_type)

    def hash_state(self, state):
        if self.uses_true_hash:
            _hash = hash(
                tuple(
                    [
                        self.hash_tensor(x) if k in state._tensor_names else hash(x)
                        for k, x in state.items()
                    ],
                ),
            )
            return _hash
        return self.get_one_id()


hasher = Hasher()



import numpy


AVAILABLE_FUNCTIONS = [
    "abs",
    "argmax",
    "hash_numpy",
    "hash_tensor",
    "concatenate",
    "stack",
    "clip",
    "repeat",
    "min",
    "max",
    "norm",
    "unsqueeze",
    "where",
    "sqrt",
    "tile",
    "logical_or",
    "logical_and",
    "logical_not",
    "ones",
    "zeros",
    "arange",
    "full_like",
    "allclose",
]


def concatenate(x, axis=0, out=None):
    return numpy.concatenate(x, axis=axis, out=out)


def stack(x, axis=0, out=None):
    return numpy.stack(x, axis=axis, out=out)


def clip(x, a_min, a_max, out=None):
    return numpy.clip(x, a_min, a_max, out=out)


def repeat(x, repeat, axis=None):
    return numpy.repeat(x, repeat, axis=axis)


def tile(x, repeat):
    return numpy.tile(x, repeat)


def norm(x, ord=None, axis=None, keepdims=False):
    return numpy.linalg.norm(x, ord=ord, axis=axis, keepdims=keepdims)


def min(x, axis=None, other=None, out=None):
    if other is None:
        return numpy.min(x, axis=axis, out=out)
    else:
        return numpy.minimum(x, other, out=out)


def max(tensor, axis=None, other=None, out=None):
    if other is None:
        return numpy.max(tensor, axis=axis, out=out)
    else:
        return numpy.maximum(tensor, other, out=out)


def unsqueeze(x, axis=0):
    return numpy.expand_dims(x, axis=axis)


def where(cond, a, b, *args, **kwargs):
    return numpy.where(cond, a, b, *args, **kwargs)


def argmax(x, axis=None, *args, **kwargs):
    return numpy.argmax(x, axis=axis, *args, **kwargs)


def sqrt(x, *args, **kwargs):
    return numpy.sqrt(x, *args, **kwargs)


def logical_or(a, b):
    return numpy.logical_or(a, b)


def logical_and(a, b):
    return numpy.logical_and(a, b)


def logical_not(x):
    return numpy.logical_not(x)


def ones(*args, **kwargs):
    return numpy.ones(*args, **kwargs)


def zeros(*args, **kwargs):
    return numpy.zeros(*args, **kwargs)


def arange(*args, **kwargs):
    return numpy.arange(*args, **kwargs)


def abs(*args, **kwargs):
    return numpy.abs(*args, **kwargs)


def allclose(*args, **kwargs):
    return numpy.allclose(*args, **kwargs)


def full_like(*args, **kwargs):
    return numpy.full_like(*args, **kwargs)


def mod(*args, **kwargs):
    return numpy.mod(*args, **kwargs)



import pytest


pytest.skip(allow_module_level=True)
try:
    import networkx

    from judo.data_structures.tree import HistoryTree, NetworkxTree, to_node_id
    from judo.functions.api import API

    MISSING_IMPORTS = False
except ImportError:
    MISSING_IMPORTS = True


def random_powerlaw():
    g = networkx.DiGraph()
    t = networkx.random_powerlaw_tree(500, gamma=3, tries=1000, seed=160290)
    graph = networkx.compose(g, t)
    mapping = {n: to_node_id(n) for n in graph.nodes}
    return networkx.relabel_nodes(graph, mapping)


def small_tree():
    node_data = {"a": API.arange(10), "b": API.zeros(10)}
    edge_data = {"c": API.ones(10)}
    g = networkx.DiGraph()
    for i in range(8):
        g.add_node(to_node_id(i), **node_data)
    pairs = [(0, 1), (1, 2), (2, 3), (2, 4), (2, 5), (3, 6), (3, 7)]
    for a, b in pairs:
        g.add_edge(to_node_id(a), to_node_id(b), **edge_data)
    return g


@pytest.fixture(params=[random_powerlaw, small_tree], scope="function")
def tree(request):
    tree = HistoryTree()
    tree.data = request.param()
    return tree


@pytest.mark.skipif(MISSING_IMPORTS, reason="networkx not installed")
class TestNetworkxTree:
    def test_init(self, tree):
        pass

    def test_reset_graph(self, tree):
        node_data = {"miau": 2104}
        root_id = to_node_id(421)
        tree.reset_graph(root_id=root_id, node_data=node_data, epoch=0)
        assert tree.root_id == root_id
        assert isinstance(tree.data.nodes[root_id], dict)
        assert tree.data.nodes[root_id]["epoch"] == 0
        assert len(tree) == 1
        assert tree.data.nodes[root_id]["miau"] == 2104

    def test_append_leaf(self, tree):
        node_data = {"node": API.arange(10)}
        edge_data = {"edge": False}
        leaf_id = to_node_id(-421)
        epoch = 123
        tree.append_leaf(
            leaf_id=leaf_id,
            parent_id=tree.root_id,
            node_data=node_data,
            edge_data=edge_data,
            epoch=epoch,
        )
        assert (tree.data.nodes[leaf_id]["node"] == node_data["node"]).all()
        assert tree.data.nodes[leaf_id]["epoch"] == epoch
        assert tree.data.edges[(tree.root_id, leaf_id)] == edge_data
        assert leaf_id in tree.leafs
        assert tree.root_id not in tree.leafs

    def test_prune_dead_branches(self, tree):
        leafs = tree.get_leaf_nodes()
        tree.prune_dead_branches(dead_leafs=set(leafs[:2]), alive_leafs=set(leafs[2:]))
        for le in leafs[:2]:
            assert le not in tree.leafs
            assert le not in tree.data

        assert networkx.is_tree(tree.data)

    def test_get_parent(self, tree):
        node = tuple(tree.data.nodes)[1]
        child = list(tree.data.out_edges(node))[0][1]
        assert node == tree.get_parent(child)

    def test_compose_not_crashes(self, tree):
        other = NetworkxTree()
        other.data = small_tree()
        tree.compose(other)

    def test_get_path_node_ids(self, tree):
        leafs = tree.get_leaf_nodes()
        path = tree.get_path_node_ids(leafs[0])
        other_path = tree.get_branch(leafs[0])
        assert (path == other_path).all()

    def test_path_data_generator(self, tree):
        leaf = tree.get_leaf_nodes()[0]
        path = tree.get_path_node_ids(leaf)
        data = list(tree.path_data_generator(path))
        assert len(data) == len(path) - 1
        assert len(data[0]) == 2
        for x in data:
            assert isinstance(x, tuple)
            for y in x:
                assert isinstance(y, dict)

        data = list(tree.path_data_generator(path, return_children=True))
        assert len(data) == len(path) - 1
        assert len(data[0]) == 3
        for x in data:
            assert isinstance(x, tuple)
            for y in x:
                assert isinstance(y, dict)

    def test_random_nodes_generator(self, tree):
        data = list(tree.random_nodes_generator())
        assert len(data) == len(tree.data.nodes) - 1
        assert len(data[0]) == 2
        for x in data:
            assert isinstance(x, tuple)
            for y in x:
                assert isinstance(y, dict)

        data = list(tree.random_nodes_generator(return_children=True))
        assert len(data) == len(tree.data.nodes) - 1
        assert len(data[0]) == 3
        for x in data:
            assert isinstance(x, tuple)
            for y in x:
                assert isinstance(y, dict)



import pytest

import judo


try:
    import IPython

    MISSING_IPYTHON = False
except ImportError:
    MISSING_IPYTHON = True


class TestNotebook:
    @pytest.mark.skipif(MISSING_IPYTHON, reason="IPython not installed")
    def test_remove_notebook_margin_not_crashes(self):
        judo.functions.notebook.remove_notebook_margin()

    def test_running_in_ipython_not_crashes(self):
        judo.functions.notebook.running_in_ipython()



import numpy
import pytest


try:
    import torch

    TORCH_AVAILABLE = True
except ImportError:
    TORCH_AVAILABLE = False

from judo import Backend, dtype


@pytest.fixture()
def backend():
    return Backend


backend_names = ["numpy", "torch"] if TORCH_AVAILABLE else ["numpy"]


class TestDataTypes:
    @pytest.mark.parametrize("name", backend_names)
    def test_bool(self, backend, name):
        backend.set_backend(name)
        if name == "numpy":
            assert dtype.bool == numpy.bool_, dtype.bool
        elif name == "torch":
            assert dtype.bool == torch.bool



import pytest

from judo import API, Backend


@pytest.fixture()
def backend():
    return Backend


class TestAPI:
    def test_from_judo(self, backend):
        x = API.zeros((10, 10))
        assert API.sqrt(x).sum() == 0



AVAILABLE_FUNCTIONS = {"running_in_ipython", "remove_notebook_margin"}


def running_in_ipython() -> bool:
    """Return ``True`` if the code is this function is being called from an IPython kernel."""
    try:
        from IPython import get_ipython

        return get_ipython() is not None
    except ImportError:
        return False


def remove_notebook_margin(output_width_pct: int = 80):
    """Make the notebook output wider."""
    from IPython.core.display import HTML

    html = (
        "<style>"
        ".container { width:" + str(output_width_pct) + "% !important; }"
        ".input{ width:70% !important; }"
        ".text_cell{ width:70% !important;"
        " font-size: 16px;}"
        ".title {align:center !important;}"
        "</style>"
    )
    return HTML(html)



from typing import Any, Dict, Generator, List, Set, Tuple, Union

import numpy

from judo.judo_backend import torch


StateDict = Dict[str, Dict[str, Any]]

Tensor = Union[numpy.ndarray, torch.Tensor]
Vector = Union[numpy.ndarray, torch.Tensor]
Matrix = Union[numpy.ndarray, torch.Tensor]
Scalar = Union[int, float]

NodeId = Union[str, int]
NodeData = Union[Tuple[dict, dict], Tuple[dict, dict, dict]]
NodeDataGenerator = Generator[NodeData, None, None]
NamesData = Union[Tuple[str], Set[str], List[str]]



import pytest

from judo import dtype
from judo.data_structures.states import States
from judo.functions.api import API


states_classes = [States]


@pytest.fixture(scope="class", params=states_classes)
def states_class(request):
    return request.param


class TestStates:
    def test_init_dict(self, states_class):
        state_dict = {"name_1": {"size": tuple([1]), "dtype": dtype.float32}}
        new_states = states_class(state_dict=state_dict, batch_size=2)
        assert new_states.n == 2

    def test_init_kwargs(self, states_class):
        name = "miau"
        new_states = states_class(batch_size=2, miau=name)
        assert new_states._batch_size == 2
        assert name in new_states.keys()
        assert getattr(new_states, name) == name, type(new_states)

    def test_getitem(self, states_class):
        name = "miau"
        new_states = states_class(batch_size=2, miau=name)
        assert new_states[name] == name, type(new_states)

    def test_setitem(self, states_class):
        name_1 = "miau"
        val_1 = name_1
        name_2 = "elephant"
        val_2 = API.arange(10)
        new_states = states_class(batch_size=2)
        new_states[name_1] = val_1
        new_states[name_2] = val_2
        assert new_states[name_1] == val_1, type(new_states)
        assert (new_states[name_2] == val_2).all(), type(new_states)

    def test_repr(self, states_class):
        name = "miau"
        new_states = states_class(batch_size=2, miau=name)
        assert isinstance(new_states.__repr__(), str)

    def test_n(self, states_class):
        new_states = states_class(batch_size=2)
        assert new_states.n == new_states._batch_size == 2

    def test_get(self, states_class):
        new_states = states_class(batch_size=2, test="test")
        assert new_states.get("test") == "test"
        assert new_states.get("AKSJDFKG") is None
        assert new_states.get("ASSKADFKA", 5) == 5

    def test_split_states(self, states_class):
        batch_size = 20
        new_states = states_class(batch_size=batch_size, test="test")
        for s in new_states.split_states(batch_size):
            assert len(s) == 1
            assert s.test == "test"
        data = API.repeat(API.arange(5).reshape(1, -1), batch_size, 0)
        new_states = states_class(batch_size=batch_size, test="test", data=data)
        for s in new_states.split_states(batch_size):
            assert len(s) == 1
            assert s.test == "test"
            assert bool((s.data == API.arange(5)).all()), s.data
        chunk_len = 4
        test_data = API.repeat(API.arange(5).reshape(1, -1), chunk_len, 0)
        for s in new_states.split_states(5):
            assert len(s) == chunk_len
            assert s.test == "test"
            assert (s.data == test_data).all(), (s.data.shape, test_data.shape)

        batch_size = 21
        data = API.repeat(API.arange(5).reshape(1, -1), batch_size, 0)
        new_states = states_class(batch_size=batch_size, test="test", data=data)
        chunk_len = 5
        test_data = API.repeat(API.arange(5).reshape(1, -1), chunk_len, 0)
        split_states = list(new_states.split_states(5))
        for s in split_states[:-1]:
            assert len(s) == chunk_len
            assert s.test == "test"
            assert (s.data == test_data).all(), (s.data.shape, test_data.shape)

        assert len(split_states[-1]) == 1
        assert split_states[-1].test == "test"
        assert (split_states[-1].data == API.arange(5)).all(), (s.data.shape, test_data.shape)

    def test_get_params_dir(self, states_class):
        state_dict = {"name_1": {"size": tuple([1]), "dtype": dtype.float32}}
        new_states = states_class(state_dict=state_dict, batch_size=2)
        params_dict = new_states.get_params_dict()
        assert isinstance(params_dict, dict)
        for k, v in params_dict.items():
            assert isinstance(k, str)
            assert isinstance(v, dict)
            for ki, _ in v.items():
                assert isinstance(ki, str)

    def test_merge_states(self, states_class):
        batch_size = 21
        data = API.repeat(API.arange(5).reshape(1, -1), batch_size, 0)
        new_states = states_class(batch_size=batch_size, test="test", data=data)
        split_states = tuple(new_states.split_states(batch_size))
        merged = new_states.merge_states(split_states)
        assert len(merged) == batch_size
        assert merged.test == "test"
        assert (merged.data == data).all()

        split_states = tuple(new_states.split_states(5))
        merged = new_states.merge_states(split_states)
        assert len(merged) == batch_size
        assert merged.test == "test"
        assert (merged.data == data).all()



import logging
from typing import Dict, Iterable, List, Tuple, Union

import judo
from judo.functions import random_state
from judo.judo_backend import Backend
from judo.typing import Tensor


class ReplayMemory:
    """Replay buffer that contains data collected from algorithm runs."""

    _log = logging.getLogger("Memory")

    def __init__(
        self,
        max_size: int,
        names: Union[List[str], Tuple[str]],
        min_size: int = None,
    ):
        """
        Initialize a :class:`ReplayMemory`.

        Args:
            max_size: Maximum number of experiences that will be stored.
            names: Names of the replay data attributes that will be stored.
            min_size: Minimum number of samples that need to be stored before the \
                     replay memory is considered ready. If ``None`` it will be equal \
                     to max_size.

        """
        self.max_size = max_size
        self.min_size = 1.0 if min_size is None else min_size
        self.names = names
        self.reset()

    def __len__(self) -> int:
        first_attr = getattr(self, self.names[0])
        return 0 if first_attr is None else len(first_attr)

    def __repr__(self) -> str:
        text = "Memory with min_size %s max_size %s and length %s" % (
            self.min_size,
            self.max_size,
            len(self),
        )
        return text

    def reset(self):
        """Delete all the data previously stored in the memory."""
        for name in self.names:
            setattr(self, name, None)

    def get(self, name):
        """Get attributes of the memory."""
        if name == "len":
            return len(self)
        return getattr(self, name)

    def is_ready(self) -> bool:
        """
        Return ``True`` if the number of experiences in the memory is greater than ``min_size``.
        """
        return len(self) >= self.min_size

    def get_values(self) -> Tuple[Tensor, ...]:
        """Return a tuple containing the memorized data for all the saved data attributes."""
        return tuple([getattr(self, val) for val in self.names])

    def as_dict(self) -> Dict[str, Tensor]:
        return dict(zip(self.names, self.get_values()))

    def iterate_batches(self, batch_size: int, as_dict: bool = True):
        with Backend.use_backend("numpy"):
            indexes = random_state.permutation(range(len(self)))
            for i in range(0, len(self), batch_size):
                batch_ix = indexes[i : i + batch_size]  # noqa: E203
                data = tuple([getattr(self, val)[batch_ix] for val in self.names])
                if as_dict:
                    yield dict(zip(self.names, data))
                else:
                    yield data

    def iterate_values(self, randomize: bool = False) -> Iterable[Tuple[Tensor]]:
        """
        Return a generator that yields a tuple containing the data of each state \
        stored in the memory.
        """
        indexes = range(len(self))
        if randomize:
            with Backend.use_backend("numpy"):
                indexes = random_state.permutation(indexes)
        for i in indexes:
            yield tuple([getattr(self, val)[i] for val in self.names])

    def append(self, **kwargs):
        for name, val in kwargs.items():
            if name not in self.names:
                raise KeyError("%s not in self.names: %s" % (name, self.names))
            # Scalar vectors are transformed to columns
            val = judo.to_backend(val)
            if len(val.shape) == 0:
                val = judo.unsqueeze(val)
            if len(val.shape) == 1:
                val = val.reshape(-1, 1)
            try:
                processed = (
                    val
                    if getattr(self, name) is None
                    else judo.concatenate([getattr(self, name), val])
                )
                if len(processed) > self.max_size:
                    processed = processed[: self.max_size]
            except Exception as e:
                print(name, val.shape, getattr(self, name).shape)
                raise e
            setattr(self, name, processed)
        self._log.info("Memory now contains %s samples" % len(self))



from typing import Callable

import numpy

import judo
from judo.data_types import dtype
from judo.functions.random import random_state
from judo.judo_tensor import tensor
from judo.typing import Tensor


AVAILABLE_FUNCTIONS = {
    "l2_norm",
    "relativize",
    "get_alive_indexes",
    "calculate_virtual_reward",
    "calculate_clone",
    "calculate_distance",
    "fai_iteration",
    "cross_virtual_reward",
    "cross_clone",
    "cross_fai_iteration",
}


def l2_norm(x: Tensor, y: Tensor) -> Tensor:
    """Euclidean distance between two batches of points stacked across the first dimension."""
    return judo.norm(x - y, axis=1)


def relativize(x: Tensor) -> Tensor:
    """Normalize the data using a custom smoothing technique."""
    orig = x
    x = judo.astype(x, dtype.float)
    std = x.std()
    if float(std) == 0:
        return judo.ones(len(x), dtype=orig.dtype)
    standard = (x - x.mean()) / std
    with numpy.errstate(invalid="ignore", divide="ignore"):
        res = judo.where(standard > 0.0, judo.log(1.0 + standard) + 1.0, judo.exp(standard))
    return res


def get_alive_indexes(oobs: Tensor):
    """Get indexes representing random alive walkers given a vector of death conditions."""
    if judo.all(oobs):
        return judo.arange(len(oobs))
    ix = judo.logical_not(oobs).flatten()
    return random_state.choice(judo.arange(len(ix))[ix], size=len(ix), replace=ix.sum() < len(ix))


def calculate_distance(
    observs: Tensor,
    distance_function: Callable = l2_norm,
    return_compas: bool = False,
    oobs: Tensor = None,
    compas: Tensor = None,
):
    """Calculate a distance metric for each walker with respect to a random companion."""
    if compas is None:
        compas = get_alive_indexes(oobs) if oobs is not None else judo.arange(observs.shape[0])
        compas = random_state.permutation(compas)
    flattened_observs = observs.view(observs.shape[0], -1)
    distance = distance_function(flattened_observs, flattened_observs[compas])
    distance_norm = relativize(distance.flatten())
    return distance_norm if not return_compas else (distance_norm, compas)


def calculate_virtual_reward(
    observs: Tensor,
    rewards: Tensor,
    oobs: Tensor = None,
    dist_coef: float = 1.0,
    reward_coef: float = 1.0,
    other_reward: Tensor = 1.0,
    return_compas: bool = False,
    distance_function: Callable = l2_norm,
):
    """Calculate the virtual rewards given the required data."""

    compas = get_alive_indexes(oobs) if oobs is not None else judo.arange(len(rewards))
    compas = random_state.permutation(compas)
    flattened_observs = observs.reshape(len(oobs), -1)
    other_reward = other_reward.flatten() if dtype.is_tensor(other_reward) else other_reward
    distance = distance_function(flattened_observs, flattened_observs[compas])
    distance_norm = relativize(distance.flatten())
    rewards_norm = relativize(rewards)

    virtual_reward = distance_norm**dist_coef * rewards_norm**reward_coef * other_reward
    return virtual_reward.flatten() if not return_compas else (virtual_reward.flatten(), compas)


def calculate_clone(virtual_rewards: Tensor, oobs: Tensor = None, eps=1e-3):
    """Calculate the clone indexes and masks from the virtual rewards."""
    compas_ix = get_alive_indexes(oobs) if oobs is not None else judo.arange(len(virtual_rewards))
    compas_ix = random_state.permutation(compas_ix)
    vir_rew = virtual_rewards.flatten()
    clone_probs = (vir_rew[compas_ix] - vir_rew) / judo.where(vir_rew > eps, vir_rew, tensor(eps))
    will_clone = clone_probs.flatten() > random_state.random(len(clone_probs))
    return compas_ix, will_clone


def fai_iteration(
    observs: Tensor,
    rewards: Tensor,
    oobs: Tensor = None,
    dist_coef: float = 1.0,
    reward_coef: float = 1.0,
    eps=1e-8,
    other_reward: Tensor = 1.0,
):
    """Perform a FAI iteration."""
    oobs = oobs if oobs is not None else judo.zeros(rewards.shape, dtype=dtype.bool)
    virtual_reward = calculate_virtual_reward(
        observs,
        rewards,
        oobs,
        dist_coef=dist_coef,
        reward_coef=reward_coef,
        other_reward=other_reward,
    )
    compas_ix, will_clone = calculate_clone(virtual_rewards=virtual_reward, oobs=oobs, eps=eps)
    return compas_ix, will_clone


def cross_virtual_reward(
    host_observs: Tensor,
    host_rewards: Tensor,
    ext_observs: Tensor,
    ext_rewards: Tensor,
    dist_coef: float = 1.0,
    reward_coef: float = 1.0,
    return_compas: bool = False,
    distance_function: Callable = l2_norm,
):
    """Calculate the virtual rewards between two cloud of points."""
    host_observs = host_observs.reshape(len(host_rewards), -1)
    ext_observs = ext_observs.reshape(len(ext_rewards), -1)
    compas_host = random_state.permutation(judo.arange(len(host_rewards)))
    compas_ext = random_state.permutation(judo.arange(len(ext_rewards)))

    # TODO: check if it's better for the distances to be the same for host and ext
    h_dist = distance_function(host_observs, ext_observs[compas_host])
    e_dist = distance_function(ext_observs, host_observs[compas_ext])
    host_distance = relativize(h_dist.flatten())
    ext_distance = relativize(e_dist.flatten())

    host_rewards = relativize(host_rewards)
    ext_rewards = relativize(ext_rewards)

    host_vr = host_distance**dist_coef * host_rewards**reward_coef
    ext_vr = ext_distance**dist_coef * ext_rewards**reward_coef
    if return_compas:
        return (host_vr, compas_host), (ext_vr, compas_ext)
    return host_vr, ext_vr


def cross_clone(
    host_virtual_rewards: Tensor,
    ext_virtual_rewards: Tensor,
    host_oobs: Tensor = None,
    eps=1e-3,
):
    """Perform a clone operation between two different groups of points."""
    compas_ix = random_state.permutation(judo.arange(len(ext_virtual_rewards)))
    host_vr = judo.astype(host_virtual_rewards.flatten(), dtype=dtype.float32)
    ext_vr = judo.astype(ext_virtual_rewards.flatten(), dtype=dtype.float32)
    clone_probs = (ext_vr[compas_ix] - host_vr) / judo.where(
        ext_vr > eps,
        ext_vr,
        tensor(eps, dtype=dtype.float32),
    )
    will_clone = clone_probs.flatten() > random_state.random(len(clone_probs))
    if host_oobs is not None:
        will_clone[host_oobs] = True
    return compas_ix, will_clone


def cross_fai_iteration(
    host_observs: Tensor,
    host_rewards: Tensor,
    ext_observs: Tensor,
    ext_rewards: Tensor,
    host_oobs: Tensor = None,
    dist_coef: float = 1.0,
    reward_coef: float = 1.0,
    distance_function: Callable = l2_norm,
    eps: float = 1e-8,
):
    """Perform a FractalAI cloning process between two clouds of points."""
    host_vr, ext_vr = cross_virtual_reward(
        host_observs=host_observs,
        host_rewards=host_rewards,
        ext_observs=ext_observs,
        ext_rewards=ext_rewards,
        dist_coef=dist_coef,
        reward_coef=reward_coef,
        distance_function=distance_function,
        return_compas=False,
    )

    compas_ix, will_clone = cross_clone(
        host_virtual_rewards=host_vr,
        ext_virtual_rewards=ext_vr,
        host_oobs=host_oobs,
        eps=eps,
    )
    return compas_ix, will_clone



import pytest

from judo import Backend


@pytest.fixture()
def backend():
    return Backend


class TestBackend:
    def test_init(self, backend):
        pass

    def test_set_backend(self, backend):
        assert backend.is_numpy()
        backend.set_backend("torch")
        assert backend.is_torch()

        assert backend.get_device() == "cpu"
        backend.set_backend(device="cuda")
        assert backend.get_device() == "cuda"

        assert not backend.requires_grad()
        backend.set_backend(requires_grad=True)
        assert backend.requires_grad()

        backend.reset_state()
        assert backend.is_numpy()
        assert backend.get_device() == "cpu"
        assert not backend.requires_grad()



from judo.data_structures.bounds import Bounds
from judo.data_structures.memory import ReplayMemory
from judo.data_structures.states import States


try:
    from judo.data_structures.tree import HistoryTree
except ImportError:
    HistoryTree = None
